@@ -1,7 +1,7 @@
 /* File related stuff
    
    Copyright (C) 1996 Maurizio Plaza
-   		 1996,1997 Jakub Jelinek
+   		 1996,1997,1999 Jakub Jelinek
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,8 +17,11 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
+/* TODO: This file is a good candidate for rewrite from scratch */
+
 #include <ctype.h>
 #include <sys/types.h>
+#include <sys/time.h>
 #include <errno.h>
 #include "silo.h"
 typedef int FILE;
@@ -35,6 +38,7 @@ static struct {
 };
 #include "ufs.h"
 #include "romfs.h"
+#include "9660.h"
 #include <setjmp.h>
 
 static errcode_t linux_open (const char *name, int flags, io_channel * channel);
@@ -59,12 +63,12 @@ static struct struct_io_manager struct_linux_manager =
 ext2_filsys fs = 0;
 static ino_t root, cwd;
 static io_manager linux_io_manager = &struct_linux_manager;
-static do_gunzip = 0;
+static int do_gunzip = 0;
 static unsigned int *gzipped_blocks;
 static unsigned int *cur_gzipped_block;
 
 static unsigned int bs;		/* Block Size */
-static unsigned long int doff;	/* Block where partition starts */
+static unsigned long long doff;	/* Block where partition starts */
 
 static unsigned char *filebuffer;
 static unsigned char *filelimit;
@@ -75,7 +79,7 @@ static int last_blockcnt;
 static char *gunzip_buffer;
 static char *gunzip_inp;
 static char *gunzip_endbuf;
-static enum { ext2, ufs, romfs } type;
+static enum { ext2, ufs, romfs, isofs } type;
 
 extern int solaris;
 
@@ -96,13 +100,14 @@ static int read_sun_partition (int partno)
 	csum ^= *ush--;
     if (csum)
 	printf ("\nWarning: Your disklabel has wrong checksum. Use fdisk to correct it.");
-    doff = (sdl.ntrks * sdl.nsect * sdl.partitions[partno - 1].start_cylinder) << 9;
+    doff = (((unsigned long long)sdl.ntrks) * sdl.nsect * sdl.partitions[partno - 1].start_cylinder) << 9;
     return 1;
 }
 
 int sprintf (char *buf, char *fmt,...)
 {
     strcpy (buf, fmt);
+    return 0;
 }
 
 void com_err (const char *a, long i, const char *fmt,...)
@@ -129,8 +134,8 @@ static errcode_t linux_open (const char *name, int flags, io_channel * channel)
     io->read_error = 0;
     io->write_error = 0;
 
-    doff = 0;
-    if (strncmp (name, "/dev/fd0", 8)) {
+    doff = 0LL;
+    if (strncmp (name, "/dev/fd0", 8) && partitionable()) {
         partno = *(name + strlen (name) - 1) - '0';
         if (partno && !read_sun_partition (partno))
 	    return EXT2_ET_BAD_DEVICE_NAME;
@@ -141,11 +146,13 @@ static errcode_t linux_open (const char *name, int flags, io_channel * channel)
 
 static errcode_t linux_close (io_channel channel)
 {
+    return 0;
 }
 
 static errcode_t linux_set_blksize (io_channel channel, int blksize)
 {
     channel->block_size = bs = blksize;
+    return 0;
 }
 
 static errcode_t linux_read_blk (io_channel channel, unsigned long block, int count, void *data)
@@ -153,7 +160,7 @@ static errcode_t linux_read_blk (io_channel channel, unsigned long block, int co
     int size;
 
     size = (count < 0) ? -count : count * bs;
-    if (read (data, size, block * bs + doff) != size) {
+    if (read (data, size, ((unsigned long long)block) * bs + doff) != size) {
 	printf ("\nRead error on block %d\n", block);
 	return EXT2_ET_SHORT_READ;
     }
@@ -162,10 +169,12 @@ static errcode_t linux_read_blk (io_channel channel, unsigned long block, int co
 
 static errcode_t linux_write_blk (io_channel channel, unsigned long block, int count, const void *data)
 {
+    return 0;
 }
 
 static errcode_t linux_flush (io_channel channel)
 {
+    return 0;
 }
 
 static void ext2fs_error (int errcode)
@@ -211,6 +220,11 @@ static int open_romfs (char *device)
     return 1;
 }
 
+static int open_isofs (char *device)
+{
+    return !iso9660_open (device, linux_io_manager, &fs);
+}
+
 static void rotate (int freq)
 {
     static int i = 0;
@@ -224,8 +238,6 @@ static void rotate (int freq)
 static int dump_block (ext2_filsys fs, blk_t * blocknr, int blockcnt,
 		       void *private)
 {
-    size_t nbytes;
-
     if (blockcnt < 0)
 	return 0;
 
@@ -347,7 +359,7 @@ static void unget_gzip_input (void)
     }
 }
 
-static gunzipped_len = 0;
+static int gunzipped_len = 0;
 
 static int dump_finish (void)
 {
@@ -385,6 +397,57 @@ static int dump_ext2 (ino_t inode, char *filename)
     return dump_finish ();
 }
 
+static int ls_ext2_proc(struct ext2_dir_entry *dirent, int offset,
+			int blocksize, char *buf, void *private)
+{
+    struct silo_inode *sino = (struct silo_inode *)filebuffer;
+    struct ext2_inode ino;
+    unsigned char *p;
+
+    strncpy(sino->name, dirent->name, dirent->name_len & 0xFF);
+    sino->name[dirent->name_len & 0xFF] = 0;
+    if (ext2fs_read_inode(fs, dirent->inode, &ino))
+	strcpy (sino->name, "--- error ---");
+    sino->mtime = ino.i_mtime;
+    sino->size = ino.i_size;
+    sino->mode = ino.i_mode;
+    sino->uid = ino.i_uid;
+    sino->gid = ino.i_gid;
+    p = strchr (sino->name, 0) + 1;
+    if (LINUX_S_ISLNK (ino.i_mode)) {
+	if (ino.i_blocks) {
+	    if (io_channel_read_blk(fs->io, ino.i_block[0], 1, p))
+		ino.i_size = 0;
+	} else {
+	    strncpy (p, (char *)&(ino.i_block[0]),ino.i_size);
+	}
+	p[ino.i_size] = 0;
+	p += ino.i_size + 1;
+    }
+    if ((long)p & 3) p += 4 - ((long)p & 3);
+    sino->inolen = p - filebuffer;
+    filebuffer = p;
+    return 0;
+}
+
+static int ls_ext2 (ino_t inode)
+{
+    errcode_t retval;
+    struct silo_inode *sino;
+
+    retval = ext2fs_dir_iterate (fs, inode, DIRENT_FLAG_INCLUDE_EMPTY,
+				 0, ls_ext2_proc, 0);
+    if (retval) {
+        printf ("\n");
+    	ext2fs_error (retval);
+        printf ("\n");
+	return 0;
+    }
+    sino = (struct silo_inode *)filebuffer;
+    sino->inolen = 0;
+    return 1;
+}
+
 static int dump_ufs (ino_t inode, char *filename)
 {
     if (ufs_block_iterate (fs, inode, dump_block, 0)) {
@@ -403,11 +466,19 @@ static int dump_romfs (ino_t inode, char *filename)
     return dump_finish ();
 }
 
+static int dump_isofs (struct iso9660_inode *inode, char *filename)
+{
+    if (iso9660_block_iterate (fs, inode, dump_block, 0)) {
+	printf ("Error while loading of %s", filename);
+	return 0;
+    }
+    return dump_finish ();
+}
+
 int dump_device_range (char *filename, char *bogusdev, int *len, void (*lenfunc)(int, char **, char **))
 {
     /* Range of blocks on physical block device */
-    int start, end = -1;
-    int retval;
+    int start = 0, end = -1;
     char *p;
 
     bs = 512;
@@ -466,55 +537,8 @@ int dump_device_range (char *filename, char *bogusdev, int *len, void (*lenfunc)
     return 0;
 }
 
-int dump_cdrom (char *filename, char *bogusdev, int *len, void (*lenfunc)(int, char **, char **))
+static int get_len (ino_t inode)
 {
-    /* Blocks on cdrom device */
-    int start, end = -1;
-    int retval;
-    int length;
-    extern char silo_conf[256];
-
-    bs = 1024;
-    bogusdev[8] = '0';
-    retval = atoi (filename + 3);
-    start = *(unsigned int *)(silo_conf + 8 * retval + 8);
-    length = *(unsigned int *)(silo_conf + 8 * retval + 12);
-    end = start + ((length + 1023) >> 10);
-    if (lenfunc)
-        (*lenfunc)(length, (char **)&filebuffer, (char **)&filelimit);
-    fs = (ext2_filsys) malloc (sizeof (struct struct_ext2_filsys));
-    if (fs) {
-	if (!linux_open (bogusdev, 0, &fs->io)) {
-	    blk_t tmp;
-	    
-	    first_block = do_gunzip;
-	    block_no = 0;
-	    last_blockcnt = 0;
-	    block_cnt = 0;
-	    for (tmp = start; tmp < end; tmp++) {
-		if (dump_block (fs, &tmp, tmp - start, 0))
-		    break;
-	    }
-
-	    if (tmp == end && dump_finish ()) {
-	        if (len) {
-	            if (do_gunzip)
-	                *len = gunzipped_len;
-	            else
-	                *len = length;
-	        }
-		return 1;
-	    }
-	}
-    }
-    printf ("\nInternal error while loading physical blocks from device\n");
-    return 0;
-}
-
-int get_len (ino_t inode)
-{
-    if (do_gunzip)
-	return gunzipped_len;
     if (type == ext2) {
 	struct ext2_inode ei;
 	int retval;
@@ -543,18 +567,42 @@ int get_len (ino_t inode)
     	    return 0;
     	}
     	return ri.size;
+    } else if (type == isofs) {
+    	/* XXX */
+    	return 0;
     }
     return 0;
 }
 
-int load_file (char *device, int partno, char *filename, char *buffer, char *limit, int *len, int dogunzip, void (*lenfunc)(int, char **, char **))
+static char *get_archstr(void)
 {
-    ino_t inode;
+	char *p = "sun4c";
+    	    	    
+	switch (get_architecture()) {
+	case sun4: p = "sun4"; break;
+	case sun4c: p = "sun4c"; break;
+	case sun4m: p = "sun4m"; break;
+	case sun4d: p = "sun4d"; break;
+	case sun4e: p = "sun4e"; break;
+	case sun4u: p = "sun4u"; break;
+	case sun4p: p = "sun4p"; break;
+	default: break;
+	}
+	return p;
+}
+
+int load_file (char *device, int partno, char *filename, char *buffer, char *limit, int *len, int cmd, void (*lenfunc)(int, char **, char **))
+{
+    ino_t inode = 0;
+    struct iso9660_inode iso_inode;
     int retval;
+    int syspkg = 0;
     int size = -1;
     char bogusdev[] = "/dev/sdaX";
     char *bogdev;
+    void *mmark;
 
+    mark (&mmark);
     if (!device)
 	device = bootdevice;
 #if 0
@@ -571,71 +619,87 @@ int load_file (char *device, int partno, char *filename, char *buffer, char *lim
         bogusdev[8] = partno + '0';
     if (setdisk (device) < 0)
 	return 0;
-    do_gunzip = dogunzip;
+    do_gunzip = cmd & 1;
     filebuffer = buffer;
     filelimit = limit;
-    if (!strncmp (filename, "!cd", 3)) {
-    	solaris = 0;
-        return dump_cdrom (filename, bogdev, len, lenfunc);
-    }
     if (*filename == '[') {
+	if (cmd & 2) {
+		printf ("You cannot ls a device range\n");
+		return 0;
+	}
     	solaris = 0;
-	return dump_device_range (filename, bogdev, len, lenfunc);
+	retval = dump_device_range (filename, bogdev, len, lenfunc);
+	release (mmark);
+	return retval;
     }
     if (!open_ext2 (bogdev)) {
         if (!open_romfs (bogdev)) {
-            if (!open_ufs (bogdev)) {
-                fatal ("Unable to open filesystem");
-                return 0;
-            } else type = ufs;
-        } else {
+            if (!open_isofs (bogdev)) {
+		if (!open_ufs (bogdev)) {
+                    fatal ("Unable to open filesystem");
+                    release (mmark);
+                    return 0;
+                } else type = ufs;
+            } else
+                type = isofs;
+        } else
             type = romfs;
-            solaris = 0;
-        }
-    } else {
+    } else
         type = ext2;
-        solaris = 0;
-    }
+    if (type != ufs)
+    	solaris = 0;
     if (type == ext2) {
-        if (retval = ext2fs_namei_follow (fs, root, root, filename, &inode)) {
+        if ((retval = ext2fs_namei_follow (fs, root, root, filename, &inode))) {
 	    printf ("\nCannot find %s (", filename);
 	    ext2fs_error (retval);
 	    printf (")\n");
 	    ext2fs_close (fs);
+	    release (mmark);
 	    return 0;
         }
     } else if (type == romfs) {
-        if (retval = romfs_namei (fs, root, root, filename, &inode)) {
+        if (romfs_namei (fs, root, root, filename, &inode)) {
 	    printf ("\nCannot find %s\n", filename);
+	    release (mmark);
 	    return 0;
         }
+    } else if (type == isofs) {
+	if (iso9660_namei (fs, filename, &iso_inode)) {
+	    printf ("\nCannot find %s\n", filename);
+	    release (mmark);
+	    return 0;
+	}
+	inode = 1;
     } else if (type == ufs) {
     	cwd = root;
     	if (solaris) {
     	    if (!ufs_namei (fs, root, root, "/platform", &cwd)) {
-    	    	if (!ufs_namei (fs, root, cwd, get_syspackage(), &inode))
+    	    	if (!ufs_namei (fs, root, cwd, get_syspackage(), &inode)) {
     	    	    cwd = inode;
-    	    	else {
-    	    	    char *p = "sun4c";
-    	    	    
-    	    	    switch (get_architecture()) {
-    	    	    	case sun4: p = "sun4"; break;
-    	    	    	case sun4c: p = "sun4c"; break;
-    	    	    	case sun4m: p = "sun4m"; break;
-    	    	    	case sun4d: p = "sun4d"; break;
-    	    	    	case sun4e: p = "sun4e"; break;
-    	    	    	case sun4u: p = "sun4u"; break;
-    	    	    }
-    	    	    if (!ufs_namei (fs, root, cwd, p, &inode))
+    	    	    syspkg = 1;
+    	    	} else {
+    	    	    if (!ufs_namei (fs, root, cwd, get_archstr(), &inode))
     	    	        cwd = inode;
     	    	}
     	    }
     	    if (cwd != root && *filename == '/') filename++;
     	}
         if (ufs_namei (fs, root, cwd, filename, &inode)) {
-	    printf ("\nCannot find %s.", filename);
-	    ufs_close (fs);
-	    return 0;
+            if (syspkg) {
+            	syspkg = 0;
+            	ufs_namei (fs, root, root, "/platform", &cwd);
+            	if (!ufs_namei (fs, root, cwd, get_archstr(), &inode)) {
+            	    cwd = inode;
+            	    if (!ufs_namei (fs, root, cwd, filename, &inode))
+            	        syspkg = 1;
+            	}
+            }
+            if (!syspkg) {
+	        printf ("\nCannot find %s.", filename);
+	        ufs_close (fs);
+	        release (mmark);
+	        return 0;
+	    }
         }
         if (solaris) {
             ino_t sinode;
@@ -650,26 +714,53 @@ int load_file (char *device, int partno, char *filename, char *buffer, char *lim
     }
     if (lenfunc) {
         do_gunzip = 0;
-        (*lenfunc)(size = get_len (inode), (char **)&filebuffer, (char **)&filelimit);
-        do_gunzip = dogunzip;
+        if (type == isofs)
+	    size = iso_inode.size;
+	else
+	    size = get_len (inode);
+        (*lenfunc)(size, (char **)&filebuffer, (char **)&filelimit);
+        do_gunzip = cmd & 1;
     }
     first_block = do_gunzip;
     last_blockcnt = 0;
     block_no = 0;
     block_cnt = 0;
     retval = 0;
-    if (inode)
-        switch (type) {
+    if (inode) {
+	if (cmd & 2)
+	    switch (type) {
+	    case ext2: retval = ls_ext2 (inode); break;
+	    default: printf ("ls not supported outside of ext2\n"); retval = 0; break;
+#if 0
+            case ufs: retval = ls_ufs (inode); break;
+            case romfs: retval = ls_romfs (inode); break;
+            case isofs: retval = ls_isofs (&iso_inode); break;
+#endif
+	    }
+	else
+	    switch (type) {
             case ext2: retval = dump_ext2 (inode, filename); break;
             case ufs: retval = dump_ufs (inode, filename); break;
             case romfs: retval = dump_romfs (inode, filename); break;
-        }
-    if (retval && len)
-        *len = (size != -1) ? size : get_len (inode);
+            case isofs: retval = dump_isofs (&iso_inode, filename); break;
+	    }
+    }	    
+    if (retval && len) {
+	if (size != -1)
+	    *len = size;
+	else if (do_gunzip)
+	    *len = gunzipped_len;
+	else if (type == isofs)
+	    *len = iso_inode.size;
+	else
+	    *len = get_len (inode);
+    }
     switch (type) {
 	case ext2: ext2fs_close (fs); break;
 	case ufs: ufs_close (fs); break;
 	case romfs: romfs_close (fs); break;
+	case isofs: iso9660_close (fs); break;
     }
+    release (mmark);
     return retval;
 }
