@@ -4,7 +4,7 @@
    		 1996 Maurizio Plaza
    		 1996 David S. Miller
    		 1996 Miguel de Icaza
-   		 1996,1997,1998 Jakub Jelinek
+   		 1996,1997,1998,1999 Jakub Jelinek
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,6 +20,8 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
+/* TODO: This file is a good candidate for rewrite from scratch */
+
 #include "silo.h"
 #include <asm/page.h>
 #include <linux/elf.h>
@@ -30,12 +32,17 @@
 
 /* This has to be first initialized variable in main.c */
 
-extern char silo_conf[256];
-extern char silo_conf_part;
+extern unsigned char silo_conf[256];
+extern unsigned char silo_conf_part, silo_conf_parts[32], raid_dsk_number;
 
 extern int start;
 
 int useconf = 0;
+enum {
+    CMD_BOOT,
+    CMD_CAT,
+    CMD_LS
+} load_cmd;
 enum arch architecture;
 static int timer_status = 0;
 static char *initrd_start;
@@ -53,7 +60,9 @@ static int given_bootargs_by_user = 0;
 static char sol_params [512];
 static int fill_reboot_cmd = 0;
 static char other_device [512];
+static int reboot = 0;
 static int floppyswap = 0;
+unsigned int linux_end = 0;
 int other_part = -1;
 int solaris = 0;
 int other = 0;
@@ -68,16 +77,11 @@ void maintabfunc (void)
 void parse_name (char *imagename, int defpart, char **device, int *part, char **kname)
 {
     static char parsebuf[1024];
-    
+
     strcpy (parsebuf, imagename);
     imagename = parsebuf;
     *device = 0;
     *kname = 0;
-    if (!strncmp (imagename, "!cd", 3)) {
-        *kname = imagename;
-        *part = 0;
-        return;
-    }
     if (prom_vers == PROM_V0) {
         static char v0_buffer[20];
         *kname = v0_device (imagename);
@@ -206,7 +210,7 @@ void check_password(char *str)
 	passwdbuff[0] = 0;
 	cmdedit ((void (*)(void)) 0, 1);
 	printf ("\n");
-	if (!strncmp (password, passwdbuff))
+	if (!strcmp (password, passwdbuff))
 	    return;
 	if (i < 2)
 	    printf ("Wrong password. Please try again...");
@@ -215,7 +219,27 @@ void check_password(char *str)
     prom_halt();
 }
 
-int get_params (char **device, int *part, char **kname, char **params)
+char *ls_options(char *parms)
+{
+    char *p = parms;
+    ls_opt = 0;
+    if (*parms == '-') {
+	for (p = parms + 1;*p && *p != ' ';p++) {
+	    switch (*p) {
+	    case 'l': ls_opt |= LSOPT_L; break;
+	    case 't': ls_opt |= LSOPT_T; break;
+	    case 'r': ls_opt |= LSOPT_R; break;
+	    default:
+		printf ("Unhandled options to ls: `%s' - only l, t and r supported\n", parms);
+	    	break;
+	    }
+	}
+	while (*p == ' ') p++;
+    }
+    return p;
+}
+
+int get_params (char **device, int *part, char **kname, char **proll, char **params)
 {
     int defpart = -1;
     char *defdevice = 0;
@@ -224,25 +248,39 @@ int get_params (char **device, int *part, char **kname, char **params)
     char *imagename = 0, *label = 0;
     int timeout = -1, beg = 0, end;
     static char buffer[2048], buffer2[2048];
+    static char prollb[256];	/* Hmm */
     static int first = 1;
     int from_prom = 0;
     int from_siloconf = 0;
     static int no_prom_args = 0;
     int tabbedout = 0;
-    char *bootblock = 0;
-    
+
+#ifndef TFTP
+    if (raid_dsk_number) {
+	defpart = silo_conf_parts[raid_dsk_number - 1];
+	if (defpart < 1 || defpart > 8)
+	    defpart = -1;
+    }
+#endif
+    load_cmd = CMD_BOOT;    
     other = 0;
     fill_reboot_cmd = 0;
     solaris = 0;
     initrd_string = 0;
     pause_after = 0;
+    reboot = 0;
+    *proll = 0;
     cmdinit ();
     *params = "";
     if (useconf) {
 	defdevice = cfg_get_strg (0, "device");
-	p = cfg_get_strg (0, "partition");
-	if (p && *p >= '1' && *p <= '8' && !p[1])
-	    defpart = *p - '0';
+	if (defpart == -1) {
+	    /* For RAID1 avoid using this at all, it makes no sense
+	       anyway */
+	    p = cfg_get_strg (0, "partition");
+	    if (p && *p >= '1' && *p <= '8' && !p[1])
+		defpart = *p - '0';
+	}
 	if (first) {
 	    first = 0;
 	    p = cfg_get_strg (0, "timeout");
@@ -305,6 +343,8 @@ int get_params (char **device, int *part, char **kname, char **params)
 		        } else if (c >= ' ') {
 		            cbuff[0] = c;
 		            cbuff[1] = 0;
+			    if (cfg_get_flag (cbuff, "single-key"))
+				imagename = cbuff;
 		        }
 	            } else {
 		        imagename = cfg_get_default ();
@@ -321,6 +361,11 @@ int get_params (char **device, int *part, char **kname, char **params)
 	    if ((!*cbuff || (timeout > 0 && timer_status < 0)) && !tabbedout)
                 printf ("boot: ");
 	    cmdedit (maintabfunc, 0);
+	    if (*cbuff == ' ') {
+		for (p = cbuff; *p == ' '; p++);
+		q = cbuff;
+		while ((*q++ = *p++) != 0);
+	    }
 	    strcpy (given_bootargs, cbuff);
 	    given_bootargs_by_user = 1;
 	    printf ("\n");
@@ -352,9 +397,14 @@ int get_params (char **device, int *part, char **kname, char **params)
 		check_password ("To specify image arguments you need to enter p");
 	    from_siloconf = 1;
 	    defdevice = cfg_get_strg (label, "device");
-	    p = cfg_get_strg (label, "partition");
-	    if (p && *p >= '1' && *p <= '8' && !p[1])
-		defpart = *p - '0';
+#ifndef TFTP
+	    if (!raid_dsk_number)
+#endif
+	    {
+		p = cfg_get_strg (label, "partition");
+		if (p && *p >= '1' && *p <= '8' && !p[1])
+		    defpart = *p - '0';
+	    }
 	    *buffer = 0;
 	    q = buffer;
 	    if (cfg_get_flag (label, "fill-reboot-cmd"))
@@ -406,6 +456,11 @@ int get_params (char **device, int *part, char **kname, char **params)
 		check_initrd (label, defpart, defdevice);
 		if (**params)
 			strcpy (q, *params);
+		p = cfg_get_strg (label, "proll");
+		if (p && strlen (p) < sizeof(prollb)-1) {
+		    strcpy (prollb, p);
+		    *proll = prollb;
+		}
 	    }
 	    if (other) {
 	        char *oth_device = 0;
@@ -439,6 +494,8 @@ int get_params (char **device, int *part, char **kname, char **params)
 	    	else {
 	    	    static char bufx[8];
 	    	    
+	    	    if (architecture == sun4u)
+	    	    	reboot = 1;
 	    	    strcpy (bufx, "x[1-16]");
 	    	    *bufx = other_part + '0';
 	    	    defdevice = other_device;
@@ -471,6 +528,11 @@ int get_params (char **device, int *part, char **kname, char **params)
 	if (!imagename) {
             printf ("boot: ");
 	    cmdedit ((void (*)(void)) 0, 0);
+	    if (*cbuff == ' ') {
+		for (p = cbuff; *p == ' '; p++);
+		q = cbuff;
+		while ((*q++ = *p++) != 0);
+	    }
 	    strcpy (given_bootargs, cbuff);
 	    given_bootargs_by_user = 1;
 	    printf ("\n");
@@ -528,6 +590,35 @@ int get_params (char **device, int *part, char **kname, char **params)
     }
     strcpy (buffer2, imagename);
     *part = -2;
+    if (!label && (!strcmp (imagename, "cat") || !strcmp (imagename, "ls"))) {
+	if (*imagename == 'c')
+	    load_cmd = CMD_CAT;
+	else {
+	    load_cmd = CMD_LS;
+	    *params = ls_options (*params);
+	}
+	imagename = *params;
+	p = strchr (imagename, ' ');
+	if (!p)
+	    *params = "";
+	else {
+	    *p++ = 0;
+	    while (*p && *p <= ' ')
+		p++;
+	    *params = p;
+	}
+    } else if (label && (!strncmp (imagename, "cat ", 4) || !strncmp (imagename, "ls ", 3))) {
+	if (*imagename == 'c') {
+	    load_cmd = CMD_CAT;
+	    imagename += 4;
+	    while (*imagename == ' ') imagename++;
+	} else {
+	    load_cmd = CMD_LS;
+	    imagename += 3;
+	    while (*imagename == ' ') imagename++;
+	    imagename = ls_options (imagename);
+	}
+    }
     parse_name (imagename, -2, device, part, kname);
     if (!*device)
 	*device = defdevice;
@@ -565,6 +656,8 @@ int get_params (char **device, int *part, char **kname, char **params)
 	    } else {
 	    	other = 1;
 	    	*kname = strdup ("[1-16]");
+	    	if (architecture == sun4u)
+	    	    reboot = 1;
 	    	return 0;
 	    }
     	} else {
@@ -650,24 +743,180 @@ void initrd_lenfunc (int len, char **filebuffer, char **filelimit)
     *filelimit = initrd_start + ((len + 16383) & ~16383);
 }
 
-/* Here we are launched */
-int bootmain (void)
+int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
+    unsigned *pstart, char *image_name)
 {
-    unsigned off;
-    int len, image_len;
-    char *kname, *params, *device;
-    char *kernel_params;
-    int part;
+    int isfile = 0;
     union {
 	char *b;
 	struct aout_hdr *a;
 	Elf32_Ehdr *e;
 	Elf64_Ehdr *f;
     } hp;
+    unsigned off = 0;
+    int len = 0;
+    unsigned st = (unsigned int) base;
+
+    /* By this point the first sector is loaded (and the rest of */
+    /* the kernel) so we check if it is an executable file, either */
+    /* an a.out or an elf binary */
+
+    hp.b = base;
+    if (hp.a->magic == 0x01030107) {
+    	if (solaris) {
+    	    printf ("\nYour Solaris `ufsboot' is not an ELF image. Try again.\n");
+	    return -1;
+    	}
+	off = sizeof (struct aout_hdr);
+	if (image_len > hp.a->ltext + hp.a->ldata)
+	    len = hp.a->ltext + hp.a->ldata;
+	else
+	    len = image_len;
+	isfile = 1;
+    } else if (hp.e->e_ident[EI_MAG0] == ELFMAG0 &&
+		hp.e->e_ident[EI_MAG1] == ELFMAG1 &&
+		hp.e->e_ident[EI_MAG2] == ELFMAG2 &&
+		hp.e->e_ident[EI_MAG3] == ELFMAG3) {
+
+        if (hp.e->e_ident[EI_DATA] != ELFDATA2MSB) {
+	    fatal ("Image is not a MSB ELF");
+	    prom_halt ();
+	}
+	if (hp.e->e_ident[EI_CLASS] == ELFCLASS32) {
+	    Elf32_Phdr *p;
+
+	    p = (Elf32_Phdr *) (hp.b + hp.e->e_phoff);
+	    if (p->p_type != PT_LOAD) {
+		fatal ("Cannot find a loadable segment in your ELF image");
+		prom_halt ();
+	    }
+	    if (solaris) {
+	        int i;
+	        unsigned long sa = (unsigned long)&_start;
+
+	        for (i = 0; i < hp.e->e_phnum; i++, p++) {
+	            if (p->p_vaddr < 0x4000 + image_len ||
+				p->p_vaddr + p->p_memsz >= sa) {
+	        	fatal("Unable to handle your Solaris `ufsboot' bootloader.");
+	            	prom_halt ();
+	            }
+	            memcpy ((char *)p->p_vaddr,
+				(char *)(0x4000 + p->p_offset), p->p_filesz);
+	            if (p->p_filesz < p->p_memsz)
+	                memset ((char *)(p->p_vaddr + p->p_filesz), 0,
+				p->p_memsz - p->p_filesz);
+	        }
+	        isfile = 1;
+	        st = hp.e->e_entry;
+	    } else {
+	            int i;
+	            unsigned long n;
+
+	            Elf32_Phdr *q = p + 1;
+		            
+	            for (i = 1; i < hp.e->e_phnum; i++, q++) {
+	                if (q->p_type != PT_LOAD)
+	                    break;
+	                n = q->p_offset - p->p_offset;
+			if (q->p_vaddr - p->p_vaddr == n &&
+	                    q->p_paddr - p->p_paddr == n &&
+			    p->p_memsz == p->p_filesz &&
+	                    p->p_memsz <= n) {
+				p->p_filesz = n + q->p_filesz;
+				p->p_memsz = n + q->p_memsz;
+	                } else {
+	                    fatal("Multiple loadable segments in your ELF image");
+	                    prom_halt();
+	                }
+	            }
+	    	    off = p->p_offset + hp.e->e_entry - p->p_vaddr;
+	    	    len = p->p_filesz;
+	    	    if (len > image_len) len = image_len;
+	    	    isfile = 1;
+	    }
+	} else if (hp.e->e_ident[EI_CLASS] == ELFCLASS64) {
+	    Elf64_Phdr *p;
+	    unsigned long long n;
+	    int i;
+	    Elf64_Phdr *q;
+
+	    p = (Elf64_Phdr *) (hp.b + hp.f->e_phoff);
+	    if (p->p_type != PT_LOAD) {
+		fatal ("Cannot find a loadable segment in your ELF image");
+		prom_halt ();
+	    }
+	    if (solaris) {
+		int i;
+		unsigned long sa = (unsigned long)&_start;
+
+		for (i = 0; i < hp.f->e_phnum; i++, p++) {
+		    if (p->p_vaddr < 0x4000 + image_len ||
+			 p->p_vaddr + p->p_memsz >= sa) {
+
+			 fatal("Unable to handle your Solaris `ufsboot' bootloader.");
+			 prom_halt ();
+ 		    }
+		    memcpy ((Elf64_Addr *)(long)(p->p_vaddr),
+				(Elf64_Addr *)(long)(0x4000 + p->p_offset),
+				p->p_filesz);
+		    if (p->p_filesz < p->p_memsz)
+			memset ((Elf64_Addr *)(long)(p->p_vaddr + p->p_filesz),
+				0, p->p_memsz - p->p_filesz);
+		}
+		isfile = 1;
+		st = hp.f->e_entry;
+	    } else {
+		    q = p + 1;
+		    for (i = 1; i < hp.f->e_phnum; i++, q++) {
+			if (q->p_type != PT_LOAD)
+			    break;
+			n = q->p_offset - p->p_offset;
+			if (q->p_vaddr - p->p_vaddr == n &&
+			    q->p_paddr - p->p_paddr == n &&
+			    p->p_memsz == p->p_filesz &&
+			    p->p_memsz <= n) {
+
+			    p->p_filesz = n + q->p_filesz;
+			    p->p_memsz = n + q->p_memsz;
+			} else {
+			    fatal("Multiple loadable segments in your ELF image");
+			    prom_halt();
+			}
+		    }
+	    }
+	    off = p->p_offset + hp.f->e_entry - p->p_vaddr;
+	    len = p->p_filesz;
+	    if (len > image_len) len = image_len;
+	    isfile = 1;
+	}
+    }
+
+    if (isfile) {
+	if (pstart != NULL)
+	    *pstart = st;
+	*poff = off;
+	*plen = len;
+    } else {
+	printf ("\nUnknown image %s format\n", image_name);
+    }
+    return isfile;
+}
+
+/* Here we are launched */
+int bootmain (void)
+{
+    unsigned off = 0;
+    int len = 0, image_len;
+    char *kname, *params, *device;
+    char *proll = 0;
+    char *image_base = (char *) 0x4000;
+    char *kernel_params;
+    int part;
     int isfile, fileok = 0;
     char *p;
-    unsigned int ret_offset;
+    unsigned int ret_offset = 0;
     char *params_device = 0;
+    int silo_conf_partition;
 
     prom_ranges_init ();
     get_idprom();
@@ -676,9 +925,11 @@ int bootmain (void)
     strcpy (my_bootargs, get_bootargs (0));
 
     printf (" ");
-   
+
+#ifndef TFTP
     if (diskinit () == -1)
 	prom_halt ();
+#endif
 	
     if (architecture == sun4u) {
     	char buffer[512], *p;
@@ -698,10 +949,15 @@ int bootmain (void)
     }
 
 /* Here should be code #ifdef TFTP, that will handle loading of silo.conf via tftp and this code should be on the other side #ifndef TFTP */
-    if (*silo_conf && silo_conf_part >= 1 && silo_conf_part <= 8) {
+    silo_conf_partition = silo_conf_part;
+#ifndef TFTP
+    if (raid_dsk_number)
+	silo_conf_partition = silo_conf_parts[raid_dsk_number - 1];
+#endif
+    if (*silo_conf && silo_conf_partition >= 1 && silo_conf_partition <= 8) {
 	int len;
 	solaris = 0;
-	fileok = load_file (0, silo_conf_part, silo_conf, (unsigned char *) 0x4000, (unsigned char *) &start, &len, 1, 0);
+	fileok = load_file (0, silo_conf_partition, silo_conf, (unsigned char *) 0x4000, (unsigned char *) &start, &len, 1, 0);
 	if (!fileok || (unsigned) len >= 65535)
 	    printf ("\nCouldn't load %s\n", silo_conf);
 	else {
@@ -709,7 +965,10 @@ int bootmain (void)
 		char *p, *q;
 		int len = 0;
 		int defpart = -1;
-		int i;
+#ifndef TFTP
+		if (raid_dsk_number)
+		    defpart = silo_conf_partition;
+#endif		
 		p = cfg_get_strg (0, "message");
 		if (p) {
 		    q = cfg_get_strg (0, "partition");
@@ -734,7 +993,8 @@ int bootmain (void)
 		    password = 0;
 		}
 	    } else {
-	        printf ("Syntax error in %s\nPlease check the file for obvious bugs, and if you think it is correct, get"
+	        printf ("Syntax error in %s\n"
+			"Please check the file for obvious bugs, and if you think it is correct, get\n"
 	                "latest version of SILO. If problems still survive, contact the author\n", silo_conf);
 	    }
 	}
@@ -748,7 +1008,7 @@ int bootmain (void)
 
     isfile = 0;			/* RC = 0 invalid file or not an executable */
     while (!isfile) {
-	switch (get_params (&device, &part, &kname, &params)) {
+	switch (get_params (&device, &part, &kname, &proll, &params)) {
 	case 1:
 	    prom_halt ();
 	case 2:
@@ -757,6 +1017,9 @@ int bootmain (void)
 	}
 	if (!kname)
 	    continue;
+	    
+	if (other && reboot)
+	    break;
 
 	if (solaris) {
 	    char *p = seed_part_into_device ((!device || !*device) ? bootdevice : device, part);
@@ -768,124 +1031,55 @@ int bootmain (void)
 	        strcat (params_device, params);
 	    }
 	}
-	    
-	fileok = load_file (device, part, kname, (unsigned char *) 0x4000, (unsigned char *) &start, &image_len, 1, 0);
-	
-	ret_offset = 0x4000;
 
-	if (fileok) {
-	    /* By this point the first sector is loaded (and the rest of */
-	    /* the kernel) so we check if it is an executable file, either */
-	    /* an a.out or an elf binary */
-
-	    hp.b = (char *) 0x4000;
-	    if (hp.a->magic == 0x01030107) {
-	    	if (solaris) {
-	    	    printf ("\nYour Solaris `ufsboot' is not an ELF image. Try again.\n");
-	    	    continue;
-	    	}
-		off = sizeof (struct aout_hdr);
-		if (image_len > hp.a->ltext + hp.a->ldata)
-		    len = hp.a->ltext + hp.a->ldata;
-		else
-		    len = image_len;
-		isfile = 1;
-	    } else {
-		if (hp.e->e_ident[EI_MAG0] == ELFMAG0 &&
-		    hp.e->e_ident[EI_MAG1] == ELFMAG1 &&
-		    hp.e->e_ident[EI_MAG2] == ELFMAG2 &&
-		    hp.e->e_ident[EI_MAG3] == ELFMAG3) {
-		    unsigned int e_entry;
-		    
-		    if (hp.e->e_ident[EI_DATA] != ELFDATA2MSB) {
-			fatal ("Image is not a MSB ELF");
-			prom_halt ();
-		    }
-		    if (hp.e->e_ident[EI_CLASS] == ELFCLASS32) {
-			Elf32_Phdr *p;
-			
-		        p = (Elf32_Phdr *) (hp.b + hp.e->e_phoff);
-		        if (p->p_type != PT_LOAD) {
-			    fatal ("Cannot find a loadable segment in your ELF image");
-			    prom_halt ();
-		        }
-		        if (solaris) {
-		            int i;
-		            unsigned long sa = (unsigned long)&_start;
-		        
-		            for (i = 0; i < hp.e->e_phnum; i++, p++) {
-		                if (p->p_vaddr < 0x4000 + image_len || p->p_vaddr + p->p_memsz >= sa) {
-		            	    fatal("Unable to handle your Solaris `ufsboot' bootloader.");
-		            	    prom_halt ();
-		                }
-		                memcpy ((char *)p->p_vaddr, (char *)(0x4000 + p->p_offset), p->p_filesz);
-		                if (p->p_filesz < p->p_memsz)
-		            	    memset ((char *)(p->p_vaddr + p->p_filesz), 0, p->p_memsz - p->p_filesz);
-		            }
-		            isfile = 1;
-		            ret_offset = hp.e->e_entry;
-		        } else {
-		            int i;
-		            unsigned long n;
-		            Elf32_Phdr *q = p + 1;
-		            
-		            for (i = 1; i < hp.e->e_phnum; i++, q++) {
-		                if (q->p_type != PT_LOAD)
-		                    break;
-		                n = q->p_offset - p->p_offset;
-				if (q->p_vaddr - p->p_vaddr == n &&
-		                    q->p_paddr - p->p_paddr == n &&
-				    p->p_memsz == p->p_filesz &&
-		                    p->p_memsz <= n) {
-					p->p_filesz = n + q->p_filesz;
-					p->p_memsz = n + q->p_memsz;
-		                } else {
-		                    fatal("Multiple loadable segments in your ELF image");
-		                    prom_halt();
-		                }
-		            }
-		    	    off = p->p_offset + hp.e->e_entry - p->p_vaddr;
-		    	    len = p->p_filesz;
-		    	    if (len > image_len) len = image_len;
-		    	    isfile = 1;
-		        }
-		    } else if (hp.e->e_ident[EI_CLASS] == ELFCLASS64) {
-			Elf64_Phdr *p;
-			unsigned long long n;
-			int i;
-			Elf64_Phdr *q;
-			
-		        p = (Elf64_Phdr *) (hp.b + hp.f->e_phoff);
-		        if (p->p_type != PT_LOAD) {
-			    fatal ("Cannot find a loadable segment in your ELF image");
-			    prom_halt ();
-		        }
-		        q = p + 1;
-		        for (i = 1; i < hp.f->e_phnum; i++, q++) {
-		                if (q->p_type != PT_LOAD)
-		                    break;
-		                n = q->p_offset - p->p_offset;
-				if (q->p_vaddr - p->p_vaddr == n &&
-		                    q->p_paddr - p->p_paddr == n &&
-				    p->p_memsz == p->p_filesz &&
-		                    p->p_memsz <= n) {
-					p->p_filesz = n + q->p_filesz;
-					p->p_memsz = n + q->p_memsz;
-		                } else {
-		                    fatal("Multiple loadable segments in your ELF image");
-		                    prom_halt();
-		                }
-		        }
-		    	off = p->p_offset + hp.f->e_entry - p->p_vaddr;
-		    	len = p->p_filesz;
-		    	if (len > image_len) len = image_len;
-		    	isfile = 1;
-		    }
-		} else
-		    printf ("\nUnknown %s image format\n", kname);
+	if (load_cmd == CMD_BOOT && proll != 0) {
+	    if (other || solaris) {
+		printf ("\nNeither \"other\" nor \"solaris\" are compatible with proll\n");
+		continue;
 	    }
-	} else
-	    printf ("\nImage not found.... try again\n");
+	    if (!load_file (device, part, proll, (unsigned char *) 0x4000,
+			(unsigned char *) 0x40000, &image_len, 1, 0)) {
+	        printf ("\nProll not found.... try again\n");
+		continue;
+	    }
+            if (!parse_executable ((char *)0x4000, image_len, &off, &len, NULL, proll))
+		continue;
+            memcpy ((char *) 0x4000, ((char *) 0x4000) + off, len);
+
+	    image_base = (char *) 0x40000;
+	    if (!load_file (device, part, kname, image_base,
+			(unsigned char *) &start, &image_len, 1, 0)) {
+	        printf ("\nImage not found.... try again\n");
+		continue;
+	    }
+            if (!(isfile = parse_executable (image_base, image_len,
+			&off, &len, NULL, kname)))
+		continue;
+	    ret_offset = 0x4000;
+
+	} else {
+	    image_base = (unsigned char *) 0x4000;
+	    if (!load_file (device, part, kname, image_base,
+			(unsigned char *) &start, &image_len,
+			load_cmd == CMD_LS ? 2 : 1, 0)) {
+		printf ("\nImage not found.... try again\n");
+        	continue;
+            }
+
+	    if (load_cmd == CMD_CAT) {
+		*(image_base + image_len) = 0;
+		printf ("%s", image_base);
+		continue;
+	    }
+
+	    if (load_cmd == CMD_LS) {
+		do_ls (image_base);
+        	continue;
+	    }
+
+            isfile = parse_executable (image_base, image_len, &off, &len,
+		&ret_offset, kname);
+	}
     }
 
     kernel_params = 0;
@@ -894,8 +1088,8 @@ int bootmain (void)
     	params_device = sol_params;
     } else if (!other) {
     	params_device = 0;
-        memcpy ((unsigned char *) 0x4000, ((unsigned char *) 0x4000) + off, len);
-        p = find_linux_HdrS (image_len);
+        memcpy (image_base, image_base + off, len);
+        p = find_linux_HdrS (image_base, image_len);
         if (p) {
             if (fill_reboot_cmd && *(unsigned short *)(p + 8) >= 0x201) { /* ie. uses reboot_command */
                 char *q = (char *)(*(unsigned int *)(p + 24)), *r;
@@ -923,7 +1117,7 @@ int bootmain (void)
             }
             if (!dig_into_params (params) && !*(unsigned short *)(p + 12)) {
                 char *s1, *s2;
-                
+
                 s1 = cfg_get_strg(0, "root");
                 if (s1) {
                     s2 = malloc(strlen(params) + 8 + strlen(s1));
@@ -933,9 +1127,12 @@ int bootmain (void)
                     params = s2;
                 }
             }
-            if (*(unsigned short *)(p + 8) >= 0x202) {
-            	kernel_params = ((char *)(*(unsigned int *)(p + 36)&0x3fffff));
-            }
+            if (*(unsigned short *)(p + 8) >= 0x202)
+            	kernel_params = (char *)(*(unsigned int *)(p + 36)&0x3fffff);
+            if (*(unsigned short *)(p + 8) >= 0x203) {
+		linux_end = *(unsigned int *)(p + 40)&0x3fffff;
+		linux_end = (linux_end + 32 + 0x1fff) & ~0x1fff;
+	    }
             /* Some UltraAX machines have /dev/fd1 floppies only. */
             if (floppyswap) {
             	char *s1;
@@ -992,8 +1189,13 @@ int bootmain (void)
     		    free (string);
     		    if (statusok) {
 	        	extern unsigned long sun4u_initrd_pa;
-	        	if (architecture == sun4u)
+	        	extern unsigned long sun4m_initrd_pa;
+	        	if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
+			    *(unsigned int *)(p + 16) = linux_end;
+			else if (architecture == sun4u)
 	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
+	            	else if (sun4m_initrd_pa)
+	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
 	            	else
 	            	    *(unsigned int *)(p + 16) = ((unsigned int)initrd_start | 0xf0000000);
 	            	*(unsigned int *)(p + 20) = initrd_size;
@@ -1005,8 +1207,13 @@ int bootmain (void)
 		        if (!initrd_device) initrd_device = initrd_defdevice;
 	        	if (load_file (initrd_device, initrd_partno, initrd_kname, (unsigned char *) 0x300000, (unsigned char *) 0x380000, 0, 0, initrd_lenfunc)) {
 	        	    extern unsigned long sun4u_initrd_pa;
-	        	    if (architecture == sun4u)
+	        	    extern unsigned long sun4m_initrd_pa;
+			    if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
+				*(unsigned int *)(p + 16) = linux_end;
+			    else if (architecture == sun4u)
 	            	        *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
+			    else if (sun4m_initrd_pa)
+				*(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
 	            	    else
 	            	        *(unsigned int *)(p + 16) = ((unsigned int)initrd_start | 0xf0000000);
 	            	    *(unsigned int *)(p + 20) = initrd_size;
@@ -1017,7 +1224,9 @@ int bootmain (void)
         }
     } else {
         char *p;
-        memcpy ((unsigned char *) 0x4000, ((unsigned char *) 0x4000) + off, len);
+        
+        if (off != 0 && !reboot)
+            memcpy (image_base, image_base + off, len);
         p = seed_part_into_device (other_device, other_part);
         strcpy (other_device, p);
         params_device = other_device;
@@ -1025,18 +1234,26 @@ int bootmain (void)
     close ();
     if (timer_status >= 1)
         close_timer ();
-    set_bootargs (params, params_device);
-    if (kernel_params) {
-    	extern char barg_out[];
-    	int len = *(unsigned int *)kernel_params;
+    if (proll) {
+	set_prollargs (params, (unsigned int)image_base, len);
+	if (show_arguments) {
+	    printf ("Arguments: \"%s\"\n");
+    	    pause_after = 1;
+	}
+    } else {
+	set_bootargs (params, params_device);
+	if (kernel_params) {
+    	    extern char barg_out[];
+    	    int len = *(unsigned int *)kernel_params;
 
-    	strncpy (kernel_params + 8, barg_out, len);
-    	kernel_params [8 + len - 1] = 0;
-    	*(unsigned int *)(kernel_params + 4) = 1;
-    }
-    if (show_arguments) {
-    	show_bootargs ();
-    	pause_after = 1;
+    	    strncpy (kernel_params + 8, barg_out, len);
+    	    kernel_params [8 + len - 1] = 0;
+    	    *(unsigned int *)(kernel_params + 4) = 1;
+	}
+	if (show_arguments) {
+    	    show_bootargs ();
+    	    pause_after = 1;
+	}
     }
     if (pause_after) {
         printf ("%s", pause_message);
@@ -1044,5 +1261,11 @@ int bootmain (void)
         printf ("\n");
     }
     memory_release();
+    if (other && reboot) {
+        strcpy (sol_params, params_device);
+        strcat (sol_params, " ");
+        strcat (sol_params, params);
+    	prom_reboot(sol_params);
+    }
     return ret_offset;
 }
