@@ -26,7 +26,7 @@ int strcmp(const char *cs,const char *ct)
 	return __res;
 }
 
-void *memset(void *s,char c,int count)
+void *memset(void *s,int c,size_t count)
 {
 	char *xs = (char *) s;
 	while (count--)
@@ -34,6 +34,11 @@ void *memset(void *s,char c,int count)
 	return s;
 }
 
+void __bzero(void *s,size_t count)
+{
+	memset(s,0,count);
+}
+
 void *memcpy(void *dest,const void *src,size_t count)
 {
 	char *tmp = (char *) dest, *s = (char *) src;
